
#include <base/SMPFunction.h>
#include <base/SMPInstr.h>
#include <interfaces/abstract/STARSFunction.h>
#include <interfaces/irdb/STARSFunction.h>
#include <interfaces/irdb/STARSInstruction.h>
#include <interfaces/irdb/STARSInterface.h>
#include <irdb-core>
#include <irdb-cfg>
//#include <bea_deprecated.hpp>

#include <list>

using namespace std;
using namespace IRDB_SDK;

// Detect push/jump pairs that were originally a call instruction
#define STARS_HANDLE_FIXED_CALLS 1

void STARS_IRDB_Function_t::IRDB_IR_Build_helper
		(
			SMPFunction *func,
			SMPInstr *CurrInst 
		)
{
#if 1
				SMPitype DataFlowType = CurrInst->GetDataFlowType();
				if ((DataFlowType == INDIR_CALL) || (DataFlowType == CALL)) {
					// See if IDA has determined the target of the call.
#if 0                           // Phase ordering: Do this later.
					CurrInst->AnalyzeCallInst(func->GetFirstFuncAddr());
#endif
					STARS_ea_t TargetAddr = CurrInst->GetCallTarget();
					bool LinkedToTarget = (STARS_BADADDR != TargetAddr);
					if (LinkedToTarget) {
						if (0 == TargetAddr) {
							SMP_msg("WARNING: Ignoring NULL call target (unreachable) at %lx\n",
								(unsigned long) CurrInst->GetAddr());
						}
						else {
							if (INDIR_CALL == DataFlowType) {
								pair<set<STARS_ea_t>::iterator, bool> InsertResult = func->IndirectCallTargets.insert(TargetAddr);
								if (InsertResult.second) {
									func->AllCallTargets.push_back(TargetAddr);
								}
							}
							else {
								(void) func->AddDirectCallTarget(TargetAddr);
							}
						}
					}
					if (DataFlowType == INDIR_CALL) {
						func->SetHasIndirectCalls();
						if (!LinkedToTarget && (!CurrInst->MDIsSystemCall())) {
							func->SetHasUnresolvedIndirectCalls();
						}
					}
				} // end if INDIR_CALL or CALL
				else if (DataFlowType == INDIR_JUMP) {
					func->IndirectJumps = true;
// was if STARS_AUDIT_INDIR_JUMP_XREFS
#if 0
					PreviousIndirJumpAddr = addr;
#endif
				}
				else if (DataFlowType == RETURN) {
					func->HasReturnInst = true;
				}
/*
				else if (CurrInst->IsBranchToFarChunk()) {
					probably doesn't apply to IRDB.
				}	
*/



#endif
}

void STARS_IRDB_Function_t::BuildFuncIR(SMPFunction *func) 
{

	cout<<"Building IR for func "<<the_func->getName()<<endl;

	typedef list<SMPInstr*> SMPInstrList_t;

	SMPInstrList_t::iterator FirstInBlock = func->Instrs.end();
	SMPInstrList_t::iterator LastInBlock = func->Instrs.end();
	bool GoodRTL=false;
	func->BuiltRTLs = true;


	// build cfg 
	auto cfgp=ControlFlowGraph_t::factory (the_func);
	auto &cfg=*cfgp;

	// sort the blocks into any order where the entry block is first.
	vector<BasicBlock_t*> sorted_blocks;

	sorted_blocks.push_back(cfg.getEntry());
	for(auto it=cfg.getBlocks().begin();
		it!=cfg.getBlocks().end(); ++it)
	{
		auto block=*it;
		if(cfg.getEntry() ==  block)	// we did the entry block first.
			continue;
		sorted_blocks.push_back(block);
	}




	bool first_in_func=true;
	for(auto block_it=sorted_blocks.begin(); block_it!=sorted_blocks.end(); ++block_it)
	{
		list<SMPInstr*> block_id_map;
		block_id_map.clear();
		auto block=*block_it;
		for(auto insn_it=block->getInstructions().begin(); insn_it!=block->getInstructions().end(); ++insn_it)
		{
			auto irdb_insn=*insn_it;
//			cout<<"Building IR for instruction "<<dec<<irdb_insn->getBaseID()<<":"<<irdb_insn->getDisassembly()<<endl;
			STARS_IRDB_Instruction_t* interface_insn=new STARS_IRDB_Instruction_t(irdb_insn);
			STARS_InstructionID_t instruction_id=interface_insn->GetID();
			// create instruction
			SMPInstr *CurrInst = new SMPInstr(interface_insn->GetID().GetIDWithinFile());
			CurrInst->Analyze();

			if(first_in_func)
			{
				first_in_func=false;

				// create SSA marker instruction
				SMPInstr *MarkerInst = new SMPInstr(STARS_SSA_MARKER_PSEUDO_ID);
				block_id_map.push_back(MarkerInst);

				MarkerInst->AnalyzeMarker();
				assert(STARS_NN_fnop == MarkerInst->GetIDAOpcode());
				GoodRTL = MarkerInst->HasGoodRTL();
				func->BuiltRTLs = (func->BuiltRTLs && GoodRTL);
				if (!GoodRTL) {
					cout << "Could not build RTLs!" << endl;
				}
				assert(FirstInBlock == func->Instrs.end());
				func->Instrs.push_back(MarkerInst);


				// find all callers.
				//for each instruction id, FromAddr, that might jump 
				//	 (but not fallthrough):
				STARS_IRDB_Interface_t *sGi=dynamic_cast<STARS_IRDB_Interface_t*>
					(global_stars_interface);
				assert(sGi);
				Instruction_t* entry=the_func->getEntryPoint();
				for(InstructionSet_t::iterator tmp_it=sGi->instruction_preds[entry].begin();
					tmp_it!=sGi->instruction_preds[entry].end(); ++tmp_it)
				{
					Instruction_t* tmp=*tmp_it;
					if(tmp->getFunction()!=the_func && tmp->getFallthrough()!=entry)
						// tell stars about the jump
						func->AddCallSource(tmp->getBaseID());
				}

				if(the_func->getEntryPoint()->getIndirectBranchTargetAddress())
					func->PossibleIndirectCallTarget = true;


			}
			block_id_map.push_back(CurrInst);
	
			// do a bunch of stuff that is based on the currently built IR.  why 
			// this needs to be specific to IRDB I don't know.  Ask clc.
			IRDB_IR_Build_helper(func,CurrInst);

			func->BuiltRTLs = (func->BuiltRTLs && CurrInst->HasGoodRTL());
//			CurrInst->Dump();

			// Insert instruction at end of list.
			func->Instrs.push_back(CurrInst);
			
		}

		// find start/end of block

		// get first and last insn in the block.
		FirstInBlock=find( func->Instrs.begin(), func->Instrs.end(), *(block_id_map.begin()));
		LastInBlock =find( func->Instrs.begin(), func->Instrs.end(), *(--block_id_map.end()));
		SMPBasicBlock *NewBlock = new SMPBasicBlock(func, FirstInBlock, LastInBlock);
		func->Blocks.push_back(NewBlock);
		func->BlockCount += 1;


	}
#if STARS_HANDLE_FIXED_CALLS
	this->FindFixedCalls(func);
#endif
}




bool STARS_IRDB_Function_t::IsInstIDInFunc(STARS_ea_t InstID)
{
	STARS_InstructionID_t id(InstID);
	
	// an ID might not yet have a corresponding STARS_Instruction_t object (or, STARS_IRDB_Instruction_t object in this case)
	// because the IR for that instruction may not yet have been built.  We check to see if the IR for the given ID
	// has been built, and if not, return that it can't be in "this" function.
	if(!id.HasCorrespondingInstructionIR())
		return false;

	const STARS_Instruction_t *stars_insn=id.GetInstruction();
	if (nullptr == stars_insn)
		return false;
	const STARS_IRDB_Instruction_t* interface_insn=dynamic_cast<const STARS_IRDB_Instruction_t*>(stars_insn);
	assert(interface_insn);
	IRDB_SDK::Instruction_t* irdb_insn=(IRDB_SDK::Instruction_t*)interface_insn->GetIRDBInstruction();
	InstructionSet_t::iterator it=the_func->getInstructions().find(irdb_insn);
	if(it==the_func->getInstructions().end())
		return false;
	return true;
}

// Analyze function, as soon as BuildFuncIR() finishes, to find pushes and jumps that came from calls that were transformed.
void STARS_IRDB_Function_t::FindFixedCalls(SMPFunction *CurrFunc) {
#if STARS_HANDLE_FIXED_CALLS
	bool FixedCallPushSeen = false;
	STARS_ea_t FixedCallPushTargetID = STARS_BADADDR;

	for (list<SMPInstr *>::iterator InstIter = CurrFunc->GetFirstInstIter(); InstIter != CurrFunc->GetLastInstIter(); ++InstIter) {


		SMPInstr *CurrInst = (*InstIter);
#if 1
		if (CurrInst->GetSTARSInstPtr()->IsJumpFromFixedCall()) 
		{
			CurrInst->SetFixedCallJump();
#if 0
			SMPitype FlowType = CurrInst->GetDataFlowType();
			if (FlowType == JUMP) {
				// If the current instruction has a jump target outside of the current func, it could be the fixed call jump
				//  that matches the fixed call push. We really should do more analysis to be sure.
				bool success = false;
				JumpTargetID = CurrInst->GetJumpTarget())
				bool ValidJumpTarget = (STARS_BADADDR != JumpTargetID);
				bool RecursiveFixedCall = (JumpTargetID == CurrFunc->GetFirstFuncAddr());
				if (ValidJumpTarget && ((!this->IsInstIDInFunc(JumpTargetID)) || RecursiveFixedCall)) {
					// Valid jump target outside of current function, or recursive call to start of function.
					CurrInst->SetFixedCallJump();
					cout << "Found fixed call jump to ID " << JumpTargetID << " from ID " << CurrInst->GetAddr() << endl;
					FixedCallPushSeen = false; // reset for detection of next occurrence
				}
			}
			else if (FlowType == INDIR_JUMP) {
				// Could be a fixed indirect call.
				CurrInst->SetFixedCallJump();
				cout << "Found fixed call indirect jump from ID " << CurrInst->GetAddr() << endl;
				FixedCallPushSeen = false; // reset for detection of next occurrence
			}
#endif
		}
#else
		CurrInst->Dump();
		STARS_ea_t InstAddr = CurrInst->GetAddr();
		if (CurrInst->IsFixedCallPush()) {
			FixedCallPushSeen = true;
			STARSOpndTypePtr PushedOpnd = CurrInst->GetPushedOpnd();
			if (PushedOpnd->IsImmedOp()) {
				bool success = false;
				STARS_InstructionID_Set_t TargetIDSet = CurrInst->GetReferencedIDs(success);
				if (success && (!TargetIDSet.empty())) {
					assert(1 == TargetIDSet.size());
					FixedCallPushTargetID = TargetIDSet.begin()->GetIDWithinFile();
					// trying to find push of return address after fixed call
					//  Corner case: Call could be at end of func, so pushed ID could be outside
					//  the func.
#if 1
					cout << "Found fixed call push at " << InstAddr << " to ID " << FixedCallPushTargetID << endl;
#else
					if (!this->IsInstIDInFunc(FixedCallPushTargetID)) { 
						FixedCallPushTargetID = STARS_BADADDR;
					}
					else {
						cout << "Found fixed call push at " << InstAddr << " to ID " << FixedCallPushTargetID << endl;
					}
#endif
				}
			}
		}
		if (FixedCallPushSeen && (STARS_BADADDR != FixedCallPushTargetID)) {
			SMPitype FlowType = CurrInst->GetDataFlowType();
			if (FlowType == JUMP) {
				// If the current instruction has a jump target outside of the current func, it could be the fixed call jump
				//  that matches the fixed call push. We really should do more analysis to be sure.
				bool success = false;
				STARS_InstructionID_Set_t TargetIDSet = CurrInst->GetReferencedIDs(success);
				if (success && (!TargetIDSet.empty())) {
					assert(1 == TargetIDSet.size());
					STARS_ea_t JumpTargetID = TargetIDSet.begin()->GetIDWithinFile();
					bool ValidJumpTarget = (STARS_BADADDR != JumpTargetID);
					bool RecursiveFixedCall = (JumpTargetID == CurrFunc->GetFirstFuncAddr());
					if (ValidJumpTarget && ((!this->IsInstIDInFunc(JumpTargetID)) || RecursiveFixedCall)) {
						// Valid jump target outside of current function, or recursive call to start of function.
						CurrInst->SetFixedCallJump();
						cout << "Found fixed call jump to ID " << JumpTargetID << " from ID " << CurrInst->GetAddr() << endl;
						FixedCallPushSeen = false; // reset for detection of next occurrence
					}
				}
			}
			else if (FlowType == INDIR_JUMP) {
				// Could be a fixed indirect call.
				CurrInst->SetFixedCallJump();
				cout << "Found fixed call indirect jump from ID " << CurrInst->GetAddr() << endl;
				FixedCallPushSeen = false; // reset for detection of next occurrence
			}
		}
#endif
	} // end for all instructions
#endif
	return;
}

// assert this represents the same function as CurrFunc
// IsIndirectCallTarget if some instruction (might) indirectly
//      branch to this function's entry.
// IsTailCallTarget if some function has a jump to this function's entry
// use IRDB-util::preds class to check that all predecessors are
// call instructions.  How to deal with push/jmp pairs?  ignore for now?
bool STARS_IRDB_Function_t::AnalyzeInstAsCallTarget(SMPFunction *CurrFunc,
		bool &IsIndirectCallTarget,
		bool &IsTailCallTarget)
{
	// Init.
	IsIndirectCallTarget=false;
	IsTailCallTarget=false;

	// Sanity
	assert(the_func);

	// Get Entry.
	Instruction_t* entry=the_func->getEntryPoint();
	assert(entry);

	// Check if entry is an IBT 
	if(entry->getIndirectBranchTargetAddress())
		// report appropriately.
		IsIndirectCallTarget=true;

	extern STARS_Interface_t* global_stars_interface;
	// safe convert.
	STARS_IRDB_Interface_t *sGi=dynamic_cast<STARS_IRDB_Interface_t*>( global_stars_interface);
	assert(sGi);


	for(InstructionSet_t::iterator it=sGi->instruction_preds[entry].begin();
		it!=sGi->instruction_preds[entry].end(); ++it)
	{
		Instruction_t* pred = *it;

		// skip any fallthroughs to the entry (weird, but could happen)
		if (pred->getFallthrough() == entry)
			continue;

		// skip any call instructions that target the branch
		const auto d=DecodedInstruction_t::factory(pred);
		if(d->isCall())
			continue;

		IsTailCallTarget=true;
	}

	return true;

}

